#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/logstream.h>
#include <deal.II/base/function.h>
#include <deal.II/base/utilities.h>
#include <deal.II/base/conditional_ostream.h>
#include <deal.II/base/work_stream.h>
#include <deal.II/base/timer.h>
#include <deal.II/base/parameter_handler.h>

#include <deal.II/lac/full_matrix.h>
#include <deal.II/lac/solver_bicgstab.h>
#include <deal.II/lac/solver_cg.h>
#include <deal.II/lac/solver_gmres.h>
#include <deal.II/lac/constraint_matrix.h>
#include <deal.II/lac/block_sparsity_pattern.h>
#include <deal.II/lac/trilinos_block_vector.h>
#include <deal.II/lac/trilinos_sparse_matrix.h>
#include <deal.II/lac/trilinos_block_sparse_matrix.h>
#include <deal.II/lac/trilinos_precondition.h>
#include <deal.II/lac/trilinos_solver.h>
#include <deal.II/lac/sparse_direct.h>

#include <deal.II/lac/precondition.h>

#include <deal.II/grid/tria.h>
#include <deal.II/grid/grid_generator.h>
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/grid/filtered_iterator.h>
#include <deal.II/grid/tria_boundary_lib.h>
#include <deal.II/grid/grid_tools.h>
#include <deal.II/grid/grid_refinement.h>

#include <deal.II/dofs/dof_handler.h>
#include <deal.II/dofs/dof_renumbering.h>
#include <deal.II/dofs/dof_accessor.h>
#include <deal.II/dofs/dof_tools.h>

#include <deal.II/fe/fe_q.h>
#include <deal.II/fe/fe_dgq.h>
#include <deal.II/fe/fe_dgp.h>
#include <deal.II/fe/fe_system.h>
#include <deal.II/fe/fe_values.h>
#include <deal.II/fe/mapping_q.h>

#include <deal.II/numerics/vectors.h>
#include <deal.II/numerics/matrices.h>
#include <deal.II/numerics/data_out.h>
#include <deal.II/numerics/error_estimator.h>
#include <deal.II/numerics/solution_transfer.h>

#include <fstream>
#include <iostream>
#include <sstream>
#include <limits>
#include <locale>
#include <string>

#include <deal.II/distributed/solution_transfer.h>
#include <deal.II/base/index_set.h>
#include <deal.II/distributed/tria.h>
#include <deal.II/distributed/grid_refinement.h>


using namespace dealii;

class IdentityPreconditioner : public Subscriptor
  {
    public:
      IdentityPreconditioner ()
    {}

      void vmult (TrilinosWrappers::MPI::Vector &dst,
                  const TrilinosWrappers::MPI::Vector &src) const
    { dst=src;}
  };

template <int dim>
class laplace_problem
{
  public:
    laplace_problem ();

    void run (int n_refs);

  private:
    void make_grid ();
    void setup_system ();
    void assemble_system ();
    void solve_direct ();
    void solve_cg ();
    void solve_amg ();
    void output_results ();


    ConditionalOStream pcout;

    parallel::distributed::Triangulation<dim>   triangulation;
    FE_Q<dim>              fe;
    DoFHandler<dim>        dof_handler;


    TrilinosWrappers::SparseMatrix system_matrix;
    ConstraintMatrix  matrix_constraints;


    TrilinosWrappers::MPI::Vector       solution;
    TrilinosWrappers::MPI::Vector      system_rhs;
    TimerOutput computing_timer;

    unsigned int n_refinement_steps;

    std_cxx1x::shared_ptr<TrilinosWrappers::PreconditionAMG>    Amg_preconditioner;
    

};


template <int dim>
laplace_problem<dim>::laplace_problem ()
		:
                pcout (std::cout,
			(Utilities::MPI::this_mpi_process(MPI_COMM_WORLD) //new in the distributed version
			 == 0)),

		triangulation (MPI_COMM_WORLD,
				typename Triangulation<dim>::MeshSmoothing
				(Triangulation<dim>::smoothing_on_refinement |
				Triangulation<dim>::smoothing_on_coarsening)),
                fe (1),
		dof_handler (triangulation),
                computing_timer (pcout,
				 TimerOutput::summary,
				 TimerOutput::wall_times)
{}


template <int dim>
void laplace_problem<dim>::make_grid ()
{

  GridGenerator::hyper_cube (triangulation, -1, 1);
  triangulation.refine_global (n_refinement_steps);

  pcout << "Number of active cells: "
	    << triangulation.n_global_active_cells()
	    << std::endl;
 

}


template <int dim>
void laplace_problem<dim>::setup_system ()
{
  dof_handler.distribute_dofs (fe);
  pcout << "Number of degrees of freedom: "
	    << dof_handler.n_dofs()
	    << std::endl;

  const unsigned int n_u = dof_handler.n_dofs();
  
  IndexSet system_index_set, system_relevant_set;
  
	//IndexSet unknowns_partitioning (n_u), unknowns_relevant_partitioning (n_u);

  system_index_set = dof_handler.locally_owned_dofs();
  
  DoFTools::extract_locally_relevant_dofs (dof_handler,
					    system_relevant_set);
  

  matrix_constraints.clear ();
  matrix_constraints.reinit (system_relevant_set);
  DoFTools::make_hanging_node_constraints (dof_handler,
                                           matrix_constraints);
  VectorTools::interpolate_boundary_values (dof_handler,
                                            0,
                                            ZeroFunction<dim>(),
                                            matrix_constraints);
  
  matrix_constraints.close ();

  TrilinosWrappers::SparsityPattern sparsity_pattern (system_index_set,
					   			MPI_COMM_WORLD);


  DoFTools::make_sparsity_pattern (dof_handler,
				     sparsity_pattern,
				     matrix_constraints, false,
				     Utilities::MPI::this_mpi_process(MPI_COMM_WORLD));
  sparsity_pattern.compress();

  system_matrix.reinit (sparsity_pattern);
  system_rhs.reinit (system_index_set, MPI_COMM_WORLD);
  solution.reinit (system_relevant_set, MPI_COMM_WORLD);
}

template <int dim>
void laplace_problem<dim>::assemble_system ()
{

  QGauss<dim>  quadrature_formula(2);

  FEValues<dim> fe_values (fe, quadrature_formula,
			 update_values | update_gradients | update_JxW_values);
  
  const unsigned int   dofs_per_cell = fe.dofs_per_cell;
  const unsigned int   n_q_points    = quadrature_formula.size();

				  
  FullMatrix<double>   cell_matrix (dofs_per_cell, dofs_per_cell);
  Vector<double>       cell_rhs (dofs_per_cell);

				 
  std::vector<unsigned int> local_dof_indices (dofs_per_cell);


  typename DoFHandler<dim>::active_cell_iterator
    cell = dof_handler.begin_active(),
    endc = dof_handler.end();
  for (; cell!=endc; ++cell)
    {
      if (cell->is_locally_owned())
      {

      fe_values.reinit (cell);

				       
      cell_matrix = 0;
      cell_rhs = 0;

				      
      for (unsigned int i=0; i<dofs_per_cell; ++i)
	for (unsigned int j=0; j<dofs_per_cell; ++j)
	  for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
	    cell_matrix(i,j) += (fe_values.shape_grad (i, q_point) *
				 fe_values.shape_grad (j, q_point) *
				 fe_values.JxW (q_point));

				      
      for (unsigned int i=0; i<dofs_per_cell; ++i)
	for (unsigned int q_point=0; q_point<n_q_points; ++q_point)
	  cell_rhs(i) += (fe_values.shape_value (i, q_point) *
			  1 *
			  fe_values.JxW (q_point));

				     
      cell->get_dof_indices (local_dof_indices);

      matrix_constraints.distribute_local_to_global(cell_matrix, local_dof_indices, system_matrix);				      
//      for (unsigned int i=0; i<dofs_per_cell; ++i)
//	for (unsigned int j=0; j<dofs_per_cell; ++j)
//	  system_matrix.add (local_dof_indices[i],
//			     local_dof_indices[j],
//			     cell_matrix(i,j));
//
      matrix_constraints.distribute_local_to_global(cell_rhs, local_dof_indices, system_rhs);			     
//      for (unsigned int i=0; i<dofs_per_cell; ++i)
//	system_rhs(local_dof_indices[i]) += cell_rhs(i);
      }
    }

				  
  system_matrix.compress();
  system_rhs.compress (Add);

}


template <int dim>
void laplace_problem<dim>::solve_direct ()
{
    TrilinosWrappers::MPI::Vector  distributed_solution(solution);
    SolverControl solver_control(system_rhs.size (), 1e-12);
    TrilinosWrappers::SolverDirect direct_solver(solver_control);
    distributed_solution.reinit(system_rhs);
    direct_solver.solve(system_matrix, distributed_solution, system_rhs);
    matrix_constraints.distribute (distributed_solution);
    solution = distributed_solution;
}

template <int dim>
void laplace_problem<dim>::solve_cg ()
{
    unsigned int cg_it=0;
    TrilinosWrappers::MPI::Vector  distributed_solution(solution);

    SolverControl solver_control (system_rhs.size(), 1e-12);
    SolverCG<TrilinosWrappers::MPI::Vector> cg (solver_control);

    distributed_solution.reinit(system_rhs);
    cg.solve (system_matrix, distributed_solution, system_rhs, IdentityPreconditioner ());

    cg_it = solver_control.last_step();
    
    matrix_constraints.distribute (distributed_solution);
    solution = distributed_solution;

    pcout<<cg_it<<"      CG iterations      "<<std::endl;
}

template <int dim>
void laplace_problem<dim>::solve_amg ()
{
    unsigned int cg_it=0;
    TrilinosWrappers::MPI::Vector  distributed_solution(solution);

    Amg_preconditioner.reset (new TrilinosWrappers::PreconditionAMG());

    std::vector<std::vector<bool> > constant_modes;
    std::vector<bool>  components (dim+1,true);
    components[dim] = false;
    DoFTools::extract_constant_modes (dof_handler, components,
                                       constant_modes);

    TrilinosWrappers::PreconditionAMG::AdditionalData Amg_data;
    Amg_data.constant_modes = constant_modes;
    Amg_data.elliptic = true;
    Amg_data.higher_order_elements = true;
    Amg_data.smoother_sweeps = 2;
    Amg_data.aggregation_threshold = 0.02;

    Amg_preconditioner->initialize(system_matrix, Amg_data);
    
    SolverControl solver_control (system_rhs.size(), 1e-12);
    SolverCG<TrilinosWrappers::MPI::Vector> cg (solver_control);

    distributed_solution.reinit(system_rhs);
    cg.solve (system_matrix, distributed_solution, system_rhs, *Amg_preconditioner);

    cg_it = solver_control.last_step();

    matrix_constraints.distribute (distributed_solution);
    solution = distributed_solution;

    pcout<<cg_it<<"      AMG iterations      "<<std::endl;
}


template <int dim>
void laplace_problem<dim>::output_results ()
{

// DataOut<dim> data_out;
//
//  data_out.attach_dof_handler (dof_handler);
//  data_out.add_data_vector (solution, "solution");
//
//  data_out.build_patches ();
//
//
//  std::ofstream output ("solution.gpl");
//  data_out.write_gnuplot (output);
    
 
  const FESystem<dim> joint_fe (fe,1);
 
  DoFHandler<dim> joint_dof_handler (triangulation);
  joint_dof_handler.distribute_dofs (joint_fe);
 
  TrilinosWrappers::MPI::Vector joint_solution;
  joint_solution.reinit (joint_dof_handler.locally_owned_dofs(), MPI_COMM_WORLD);
 
  {
    std::vector<unsigned int> local_joint_dof_indices (joint_fe.dofs_per_cell);
    std::vector<unsigned int> local_system_dof_indices (fe.dofs_per_cell);
 
    typename DoFHandler<dim>::active_cell_iterator
      joint_cell       = joint_dof_handler.begin_active(),
      joint_endc       = joint_dof_handler.end(),
      system_cell      = dof_handler.begin_active();
    for (; joint_cell!=joint_endc;
         ++joint_cell, ++system_cell)
      if (joint_cell->is_locally_owned())
        {
          joint_cell->get_dof_indices (local_joint_dof_indices);
          system_cell->get_dof_indices (local_system_dof_indices);
 
          for (unsigned int i=0; i<joint_fe.dofs_per_cell; ++i)
 
                joint_solution(local_joint_dof_indices[i])
                  = solution(local_system_dof_indices
                                  [joint_fe.system_to_base_index(i).second]);
 
        }
  }
 
 
  IndexSet locally_relevant_joint_dofs(joint_dof_handler.n_dofs());
  DoFTools::extract_locally_relevant_dofs (joint_dof_handler, locally_relevant_joint_dofs);
  TrilinosWrappers::MPI::Vector locally_relevant_joint_solution;
  locally_relevant_joint_solution.reinit (locally_relevant_joint_dofs, MPI_COMM_WORLD);
  locally_relevant_joint_solution = joint_solution;
 
  DataOut<dim> data_out;
  data_out.attach_dof_handler (joint_dof_handler);
  data_out.add_data_vector (locally_relevant_joint_solution, "solution");
  data_out.build_patches ();
 
  static int out_index=0;
  const std::string filename = ("solution-" +
                                Utilities::int_to_string (out_index, 2) +
                                "." +
                                Utilities::int_to_string
                                (triangulation.locally_owned_subdomain(), 2) +
                                ".gpl");
  std::ofstream output (filename.c_str());
  data_out.write_gnuplot (output);
  
  if (Utilities::MPI::this_mpi_process(MPI_COMM_WORLD) == 0)
    {
      
      std::string filenames;
      
      const std::string
        master_filename = ("solution-" +
                                Utilities::int_to_string (out_index, 2) +
                                ".gpl");
      std::ofstream master (master_filename.c_str());
      
      for (unsigned int i=0; i<Utilities::MPI::n_mpi_processes(MPI_COMM_WORLD); ++i)
      {
          filenames = ("solution-" +
                             Utilities::int_to_string (out_index, 2) +
                             "." +
                             Utilities::int_to_string(i, 2) +
                             ".gpl");
          
          std::ifstream infiles(filenames);
      
          master << infiles.rdbuf();
          infiles.close();
      }
      master.close();
    }

  out_index++;
  
}


template <int dim>
void laplace_problem<dim>::run (int n_refs)
{
  TrilinosWrappers::MPI::Vector      init_rhs(system_rhs);
  TrilinosWrappers::MPI::Vector      init_sol(solution);

  n_refinement_steps = n_refs;

  make_grid ();
  setup_system();
  assemble_system ();

  init_rhs = system_rhs;
  init_sol = solution;

  computing_timer.enter_section("Solve system directly");
  solve_direct ();
  computing_timer.exit_section("Solve system directly");

  solution = init_sol;
  system_rhs = init_rhs;
  computing_timer.enter_section("Solve system (CG)");
  solve_cg ();
  computing_timer.exit_section("Solve system (CG)");

  solution = init_sol;
  system_rhs = init_rhs;
  computing_timer.enter_section("Solve system (AMG)");
  solve_amg ();
  computing_timer.exit_section("Solve system (AMG)");
  output_results ();
  //computing_timer.print_summary ();
}


int main (int argc, char *argv[])
{
  Utilities::MPI::MPI_InitFinalize mpi_initialization (argc, argv);

  int n_refs;
  if (argc>=2)
    n_refs = atoi(argv[1]);
  else
    n_refs = 6;
  deallog.depth_console(0);
  laplace_problem<2> laplace_problem;
  laplace_problem.run (n_refs);

  return 0;
}
